from __future__ import print_function

import logging
import time

import pysparkling


class Manip(object):
    def __init__(self):
        self.count = 0

    def trivial_manip_with_debug(self, e):
        self.count += 1
        print('manipulating {0}'.format(e))
        return e


def test_cache_empty_partition():
    m = Manip()

    c = pysparkling.Context()
    rdd = c.parallelize(range(10), 2)
    rdd = rdd.map(m.trivial_manip_with_debug)
    rdd = rdd.filter(lambda e: e > 6).cache()
    print(rdd.collect())
    print(rdd.collect())

    print('count of map executions: {}'.format(m.count))
    assert m.count == 10


def test_timed_cache():
    m = Manip()

    # create a timed cache manager
    cm = pysparkling.TimedCacheManager(timeout=1.0)

    # create a cache entry
    c = pysparkling.Context(cache_manager=cm)
    rdd = c.parallelize(range(10), 2)
    rdd = rdd.map(m.trivial_manip_with_debug).cache()
    print(rdd.collect())
    # make sure the cache is working
    count_before = m.count
    print(rdd.collect())
    count_after = m.count
    assert count_before == count_after

    # wait to have the cache expire
    time.sleep(1.5)
    cm.gc()
    print(rdd.collect())
    assert m.count > count_after


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    # test_cache_empty_partition()
    test_timed_cache()



from operator import itemgetter


class Tokenizer(object):
    def __init__(self, expression):
        self.expression = expression

    def next(self, separator=None):
        if isinstance(separator, list):
            separator_positions_and_lengths = [
                (self.expression.find(s), s)
                for s in separator if s in self.expression
            ]
            if separator_positions_and_lengths:
                sep_pos, separator = min(separator_positions_and_lengths, key=itemgetter(0))
            else:
                sep_pos = -1
        elif separator:
            sep_pos = self.expression.find(separator)
        else:
            sep_pos = -1

        if sep_pos < 0:
            value = self.expression
            self.expression = ''
            return value

        value = self.expression[:sep_pos]
        self.expression = self.expression[sep_pos + len(separator):]
        return value


def parse_file_uri(expr):
    t = Tokenizer(expr)
    scheme = t.next('://')
    domain = t.next('/')
    last_slash_position = t.expression.rfind('/')
    folder_path = '/' + t.expression[:last_slash_position+1]
    file_pattern = t.expression[last_slash_position+1:]

    return scheme, domain, folder_path, file_pattern


def format_file_uri(scheme, domain, *local_path_components):
    return '{0}://{1}{2}'.format(scheme, domain, "/".join(local_path_components))



import tempfile

from memory_profiler import profile

import pysparkling


@profile
def main():
    tempFile = tempfile.NamedTemporaryFile(delete=True)
    tempFile.close()

    sc = pysparkling.Context()
    sc.parallelize(range(1000000)).saveAsTextFile(tempFile.name + '.gz')
    rdd = sc.textFile(tempFile.name + '.gz')
    rdd.collect()


if __name__ == '__main__':
    main()



from __future__ import absolute_import

try:
    import py7zlib
except ImportError:
    py7zlib = None

from io import BytesIO
import logging

from .codec import Codec

log = logging.getLogger(__name__)


class SevenZ(Codec):
    """Implementation of :class:`.Codec` for 7z compression.

    Needs the `pylzma` module.
    """

    def __init__(self):
        if py7zlib is None:
            log.warning('py7zlib could not be imported. To read 7z files, '
                        'install the library with "pip install pylzma".')
        super(SevenZ, self).__init__()

    def compress(self, stream):
        log.warning('Writing of 7z compressed archives is not supported.')
        return stream

    def decompress(self, stream):
        if py7zlib is None:
            return Codec.decompress(self, stream)

        uncompressed = BytesIO()

        f = py7zlib.Archive7z(file=stream)
        for f_name in f.getnames():
            uncompressed.write(f.getmember(f_name).read())

        uncompressed.seek(0)
        return uncompressed



from __future__ import division

import pysparkling


def test_mean():
    d = [1, 4, 9, 160]
    s = pysparkling.StatCounter(d)
    assert sum(d) / len(d) == s.mean()


if __name__ == '__main__':
    test_mean()



from __future__ import print_function

from collections import Counter
from contextlib import closing
import struct
import tornado.gen
import tornado.tcpclient
import tornado.testing

import pysparkling


class TCPTextTest(tornado.testing.AsyncTestCase):
    @tornado.gen.coroutine
    def client(self):
        client = tornado.tcpclient.TCPClient()
        for v in range(20):
            stream = yield client.connect('127.0.0.1', 8123)
            with closing(stream):
                stream.write('a = {}\n'.format(v).encode('utf8'))
        client.close()

    def test_connect(self):
        sc = pysparkling.Context()
        ssc = pysparkling.streaming.StreamingContext(sc, 0.1)

        counter = Counter()
        (
            ssc.socketTextStream('127.0.0.1', 8123)
            .foreachRDD(lambda rdd:
                        counter.update(''.join(rdd.collect()))
                        if rdd.collect() else None)
        )
        self.client()

        ssc.start()
        ssc.awaitTermination(timeout=0.3)
        self.assertEqual(counter['a'], 20)


class TCPBinaryFixedLengthTest(tornado.testing.AsyncTestCase):
    @tornado.gen.coroutine
    def client(self):
        client = tornado.tcpclient.TCPClient()
        stream = yield client.connect('127.0.0.1', 8124)
        with closing(stream):
            stream.write(b'hello')
        client.close()

    def test_main(self):
        sc = pysparkling.Context()
        ssc = pysparkling.streaming.StreamingContext(sc, 0.1)

        counter = Counter()
        (
            ssc.socketBinaryStream('127.0.0.1', 8124, length=5)
            .foreachRDD(lambda rdd: counter.update(rdd.collect()))
        )
        self.client()

        ssc.start()
        ssc.awaitTermination(timeout=0.3)
        self.assertEqual(counter[b'hello'], 1)


class TCPBinaryUIntLengthTest(tornado.testing.AsyncTestCase):
    @tornado.gen.coroutine
    def client(self):
        client = tornado.tcpclient.TCPClient()
        stream = yield client.connect('127.0.0.1', 8125)
        with closing(stream):
            stream.write(struct.pack('<I', 10) + b'hellohello')
        client.close()

    def test_main(self):
        sc = pysparkling.Context()
        ssc = pysparkling.streaming.StreamingContext(sc, 0.1)

        counter = Counter()
        (
            ssc.socketBinaryStream('127.0.0.1', 8125, length='<I')
            .foreachRDD(lambda rdd: counter.update(rdd.collect()))
        )
        self.client()

        ssc.start()
        ssc.awaitTermination(timeout=0.3)
        self.assertEqual(counter[b'hellohello'], 1)



"""Context."""

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from collections import defaultdict
import itertools
import logging
import pickle
import struct
import time
import traceback

from . import __version__ as PYSPARKLING_VERSION
from .broadcast import Broadcast
from .cache_manager import CacheManager
from .exceptions import ContextIsLockedException
from .fileio import File, TextFile
from .partition import Partition
from .rdd import RDD, EmptyRDD
from .task_context import TaskContext

# Python 2 compatibility
if not hasattr(time, 'perf_counter'):
    time.perf_counter = time.clock


log = logging.getLogger(__name__)


def unit_fn(arg):
    """Used as dummy serializer and deserializer."""
    return arg


def _run_task(task_context, rdd, func, partition):
    """Run a task, aka compute a partition.

    :param TaskContext task_context: this task context
    :param RDD rdd: rdd this partition is a part of
    :param func: a function
    :param Partition partition: partition to process
    """
    task_context.attempt_number += 1

    log.debug('Running stage {} for partition {} of {} (id: {}).'
              ''.format(task_context.stage_id,
                        task_context.partition_id,
                        rdd.name(), rdd.id()))

    try:
        return func(task_context, rdd.compute(partition, task_context))
    except Exception as e:  # pylint: disable=broad-except
        log.warning('Attempt {} failed for partition {} of {} (id: {}): {}'
                    ''.format(task_context.attempt_number, partition.index,
                              rdd.name(), rdd.id(), traceback.format_exc()))

        if task_context.attempt_number == task_context.max_retries:
            log.error('Partition {} of {} failed.'
                      ''.format(partition.index, rdd.name()))
            if not task_context.catch_exceptions:
                raise e

    if task_context.retry_wait:
        time.sleep(task_context.retry_wait)
    return _run_task(task_context, rdd, func, partition)


def runJob_map(i):  # pylint: disable=too-many-locals
    (deserializer, data_serializer, data_deserializer,
     serialized_func_rdd, serialized_task_context,
     serialized_data) = i

    t_start = time.perf_counter()
    func, rdd = deserializer(serialized_func_rdd)
    t_deserialize_func = time.perf_counter() - t_start

    t_start = time.perf_counter()
    partition = data_deserializer(serialized_data)
    t_deserialize_data = time.perf_counter() - t_start

    t_start = time.perf_counter()
    task_context = deserializer(serialized_task_context)
    cm_state = task_context.cache_manager.stored_idents()
    t_deserialize_task_context = time.perf_counter() - t_start

    t_start = time.perf_counter()
    result = _run_task(task_context, rdd, func, partition)
    t_exec = time.perf_counter() - t_start

    return data_serializer((
        result,
        task_context.cache_manager.get_not_in(cm_state),
        {
            'map_deserialize_func': t_deserialize_func,
            'map_deserialize_task_context': t_deserialize_task_context,
            'map_deserialize_data': t_deserialize_data,
            'map_exec': t_exec,
        }
    ))


class Context(object):
    """Context object similar to a Spark Context.

    The variable `_stats` contains measured timing information about data and
    function (de)serialization and workload execution to benchmark your jobs.

    :param pool: An instance with a ``map(func, iterable)`` method.
    :param serializer:
        Serializer for functions. Examples are `pickle.dumps` and
        `cloudpickle.dumps`.
    :param deserializer:
        Deserializer for functions. For example `pickle.loads`.
    :param data_serializer: Serializer for the data.
    :param data_deserializer: Deserializer for the data.
    :param int max_retries: maximum number a partition is retried
    :param float retry_wait: seconds to wait between retries
    :param cache_manager: custom cache manager (like `TimedCacheManager`)
    :param catch_exceptions: whether to catch and silence user space exceptions
    """

    __last_rdd_id = 0

    def __init__(self, pool=None, serializer=None, deserializer=None,
                 data_serializer=None, data_deserializer=None,
                 max_retries=3, retry_wait=0.0, cache_manager=None,
                 catch_exceptions=False):
        if pool is None:
            pool = DummyPool()
        if serializer is None:
            serializer = unit_fn
        if deserializer is None:
            deserializer = unit_fn
        if data_serializer is None:
            data_serializer = unit_fn
        if data_deserializer is None:
            data_deserializer = unit_fn
        self.max_retries = max_retries
        self.retry_wait = retry_wait

        self._cache_manager = cache_manager or CacheManager()
        self._catch_exceptions = catch_exceptions
        self._pool = pool
        self._serializer = serializer
        self._deserializer = deserializer
        self._data_serializer = data_serializer
        self._data_deserializer = data_deserializer
        self._s3_conn = None
        self._stats = defaultdict(float)
        self.locked = False

        self.version = PYSPARKLING_VERSION

    def __getstate__(self):
        r = {k: v if k not in ('_pool',) else None
             for k, v in self.__dict__.items()}
        return r

    def broadcast(self, x):
        return Broadcast(x)

    def newRddId(self):
        Context.__last_rdd_id += 1
        return Context.__last_rdd_id

    def parallelize(self, x, numSlices=None):
        """Parallelize x.

        :param x:
            An iterable (e.g. a list) that represents the data.

        :param int numSlices:
            The number of partitions the data should be split into.
            A partition is a unit of data that is processed at a time.

        :rtype: RDD
        """
        if numSlices is None or numSlices <= 1:
            return RDD([Partition(x, 0)], self)

        x_len_iter, x = itertools.tee(x, 2)
        len_x = sum(1 for _ in x_len_iter)

        def partitioned():
            for i in range(numSlices):
                start = int(i * len_x / numSlices)
                end = int((i + 1) * len_x / numSlices)
                if i + 1 == numSlices:
                    end += 1
                yield itertools.islice(x, end - start)

        return self._parallelize_partitions(partitioned())

    def _parallelize_partitions(self, partitions):
        """Helper to parallelize partitions.

        :param partitions: An iterable over the partitioned data.
        :rtype: RDD
        """

        return RDD(
            (Partition(p_data, i) for i, p_data in enumerate(partitions)),
            self,
        )

    def pickleFile(self, name, minPartitions=None):
        """Read a pickle file.

        Reads files created with :func:`RDD.saveAsPickleFile()` into an RDD.

        :param name:
            Location of a file. Can include schemes like ``http://``,
            ``s3://`` and ``file://``, wildcard characters ``?`` and ``*``
            and multiple expressions separated by ``,``.

        :param minPartitions: (optional)
            By default, every file is a partition, but this option allows to
            split these further.

        :rtype: RDD


        Example with a serialized list:

        >>> import pickle
        >>> from pysparkling import Context
        >>> from tempfile import NamedTemporaryFile
        >>> tmpFile = NamedTemporaryFile(delete=True)
        >>> tmpFile.close()
        >>> with open(tmpFile.name, 'wb') as f:
        ...     pickle.dump(['hello', 'world'], f)
        >>> Context().pickleFile(tmpFile.name).collect()[0] == 'hello'
        True

        """
        resolved_names = File.resolve_filenames(name)
        log.debug('pickleFile() resolved "{0}" to {1} files.'
                  ''.format(name, len(resolved_names)))

        n_partitions = len(resolved_names)
        if minPartitions and minPartitions > n_partitions:
            n_partitions = minPartitions

        rdd_filenames = self.parallelize(sorted(resolved_names), n_partitions)
        rdd = rdd_filenames.flatMap(
            lambda f_name: pickle.load(File(f_name).load())
        )
        rdd._name = name
        return rdd

    def runJob(self, rdd, func, partitions=None, allowLocal=False,
               resultHandler=None):
        """This function is used by methods in the RDD.

        Note that the maps are only inside generators and the resultHandler
        needs to take care of executing the ones that it needs. In other words,
        if you need everything to be executed, the resultHandler needs to be
        at least ``lambda x: list(x)`` to trigger execution of the generators.

        :param func: Map function with signature
            func(TaskContext, Iterator over elements).
        :param partitions: List of partitions that are involved. `None` means
            the map job is applied to all partitions.
        :param allowLocal: Allows local execution.
        :param resultHandler: Process the result from the maps.
        :returns: Result of resultHandler.
        :rtype: list
        """
        if not partitions:
            partitions = rdd.partitions()

        # acquire lock
        if self.locked:
            raise ContextIsLockedException
        self.locked = True

        # this is the place to insert proper schedulers
        if allowLocal or isinstance(self._pool, DummyPool):
            map_result = self._runJob_local(rdd, func, partitions)
        else:
            map_result = self._runJob_distributed(rdd, func, partitions)

        result = (resultHandler(map_result) if resultHandler is not None
                  else list(map_result))

        # release lock
        self.locked = False

        return result

    def _runJob_local(self, rdd, func, partitions):
        for partition in partitions:
            task_context = TaskContext(
                cache_manager=self._cache_manager,
                catch_exceptions=self._catch_exceptions,
                stage_id=0,
                partition_id=partition.index,
                max_retries=self.max_retries,
                retry_wait=self.retry_wait,
            )
            yield _run_task(task_context, rdd, func, partition)

    def _runJob_distributed(self, rdd, func, partitions):
        serialized_func_rdd = self._serializer((func, rdd))

        def prepare(partition):
            t_start = time.perf_counter()
            cm_clone = self._cache_manager.clone_contains(
                lambda i: i[1] == partition.index)
            self._stats['driver_cache_clone'] += (time.perf_counter() -
                                                  t_start)

            t_start = time.perf_counter()
            task_context = TaskContext(
                cache_manager=cm_clone,
                catch_exceptions=self._catch_exceptions,
                stage_id=0,
                partition_id=partition.index,
                max_retries=self.max_retries,
                retry_wait=self.retry_wait,
            )
            serialized_task_context = self._serializer(task_context)
            self._stats['driver_serialize_task_context'] += (time.perf_counter() -
                                                             t_start)

            t_start = time.perf_counter()
            serialized_partition = self._data_deserializer(partition)
            self._stats['driver_serialize_data'] += (time.perf_counter() -
                                                     t_start)

            return (
                self._deserializer,
                self._data_serializer,
                self._data_deserializer,
                serialized_func_rdd,
                serialized_task_context,
                serialized_partition,
            )

        prepared_partitions = (prepare(p) for p in partitions)
        for d in self._pool.map(runJob_map, prepared_partitions):
            t_start = time.perf_counter()
            map_result, cache_result, s = self._data_deserializer(d)
            self._stats['driver_deserialize_data'] += (time.perf_counter() -
                                                       t_start)

            # join cache
            t_start = time.perf_counter()
            self._cache_manager.join(cache_result)
            self._stats['driver_cache_join'] += time.perf_counter() - t_start

            # collect stats
            for k, v in s.items():
                self._stats[k] += v

            yield map_result

    def binaryFiles(self, path, minPartitions=None):
        """Read a binary file into an RDD.

        :param path:
            Location of a file. Can include schemes like ``http://``,
            ``s3://`` and ``file://``, wildcard characters ``?`` and ``*``
            and multiple expressions separated by ``,``.

        :param minPartitions: (optional)
            By default, every file is a partition, but this option allows to
            split these further.

        :rtype: RDD

        .. warning::
            Not part of PySpark API.


        Setting up examples:

        >>> import os, pysparkling
        >>> from backports import tempfile
        >>> sc = pysparkling.Context()
        >>> decode = lambda bstring: bstring.decode()


        Example with whole file:

        >>> with tempfile.TemporaryDirectory() as tmp:
        ...     with open(os.path.join(tmp, 'test.b'), 'wb') as f:
        ...         _ = f.write(b'bellobello')
        ...     sc.binaryFiles(tmp+'*').values().map(decode).collect()
        ['bellobello']
        """
        resolved_names = File.resolve_filenames(path)
        log.debug('binaryFile() resolved "{0}" to {1} files.'
                  ''.format(path, len(resolved_names)))

        n_partitions = len(resolved_names)
        if minPartitions and minPartitions > n_partitions:
            n_partitions = minPartitions

        rdd_filenames = self.parallelize(sorted(resolved_names), n_partitions)
        rdd = rdd_filenames.map(lambda f_name:
                                (f_name, File(f_name).load().read()))
        rdd._name = path
        return rdd

    def binaryRecords(self, path, recordLength=None):
        """Read a binary file into an RDD.

        :param path:
            Location of a file. Can include schemes like ``http://``,
            ``s3://`` and ``file://``, wildcard characters ``?`` and ``*``
            and multiple expressions separated by ``,``.

        :param recordLength:
            If `None` every file is a record, ``int`` means fixed length
            records and a ``string`` is used as a format string to ``struct``
            to read the length of variable length binary records.

        :rtype: RDD

        .. warning::
            Only an ``int`` recordLength is part of the PySpark API.


        Setting up examples:

        >>> import os, pysparkling
        >>> from backports import tempfile
        >>> sc = pysparkling.Context()
        >>> decode = lambda bstring: bstring.decode()


        Example with whole file:

        >>> with tempfile.TemporaryDirectory() as tmp:
        ...     with open(os.path.join(tmp, 'test.b'), 'wb') as f:
        ...         _ = f.write(b'bellobello')
        ...     sc.binaryRecords(tmp+'*').map(decode).collect()
        ['bellobello']


        Example with fixed length records:

        >>> with tempfile.TemporaryDirectory() as tmp:
        ...     with open(os.path.join(tmp, 'test.b'), 'wb') as f:
        ...         _ = f.write(b'bellobello')
        ...     sc.binaryRecords(tmp+'*', recordLength=5).map(decode).collect()
        ['bello', 'bello']


        Example with variable length records:

        >>> with tempfile.TemporaryDirectory() as tmp:
        ...     with open(os.path.join(tmp, 'test.b'), 'wb') as f:
        ...         _ = f.write(struct.pack('<I', 5) + b'bello')
        ...         _ = f.write(struct.pack('<I', 10) + b'bellobello')
        ...     (sc.binaryRecords(tmp+'*', recordLength='<I')
        ...      .map(decode).collect())
        ['bello', 'bellobello']
        """

        rdd = self.binaryFiles(path).values()
        if recordLength is None:
            pass
        elif isinstance(recordLength, int):
            chunker = FixedLengthChunker(recordLength)
            rdd = rdd.flatMap(chunker)
        else:
            chunker = VariableLengthChunker(recordLength)
            rdd = rdd.flatMap(chunker)
        rdd._name = path
        return rdd

    def textFile(self, filename, minPartitions=None, use_unicode=True):
        """Read a text file into an RDD.

        :param filename:
            Location of a file. Can include schemes like ``http://``,
            ``s3://`` and ``file://``, wildcard characters ``?`` and ``*``
            and multiple expressions separated by ``,``.

        :param minPartitions: (optional)
            By default, every file is a partition, but this option allows to
            split these further.

        :param use_unicode: (optional, default=True)
            Use ``utf8`` if ``True`` and ``ascii`` if ``False``.

        :rtype: RDD
        """
        resolved_names = TextFile.resolve_filenames(filename)
        log.debug('textFile() resolved "{0}" to {1} files.'
                  ''.format(filename, len(resolved_names)))

        n_partitions = len(resolved_names)
        if minPartitions and minPartitions > n_partitions:
            n_partitions = minPartitions

        encoding = 'utf8' if use_unicode else 'ascii'

        rdd_filenames = self.parallelize(sorted(resolved_names), n_partitions)
        rdd = rdd_filenames.flatMap(
            lambda f_name:
            TextFile(f_name).load(encoding=encoding).read().splitlines()
        )
        rdd._name = filename
        return rdd

    def union(self, rdds):
        """Create a union of rdds.

        :param rdds: Iterable of RDDs.
        :rtype: RDD
        """
        if all(isinstance(rdd, EmptyRDD) for rdd in rdds):
            return EmptyRDD(self)

        return self.parallelize(
            (x for rdd in rdds for x in rdd.collect())
        )

    def wholeTextFiles(self, path, minPartitions=None, use_unicode=True):
        """Read text files into an RDD of pairs of file name and file content.

        :param path:
            Location of the files. Can include schemes like ``http://``,
            ``s3://`` and ``file://``, wildcard characters ``?`` and ``*``
            and multiple expressions separated by ``,``.

        :param minPartitions: (optional)
            By default, every file is a partition, but this option allows to
            split these further.

        :param use_unicode: (optional, default=True)
            Use ``utf8`` if ``True`` and ``ascii`` if ``False``.

        :rtype: RDD
        """
        resolved_names = TextFile.resolve_filenames(path)
        log.debug('wholeTextFiles() resolved "{0}" to {1} files.'
                  ''.format(path, len(resolved_names)))

        n_partitions = len(resolved_names)
        if minPartitions and minPartitions > n_partitions:
            n_partitions = minPartitions

        encoding = 'utf8' if use_unicode else 'ascii'
        rdd_filenames = self.parallelize(
            [(f_name, encoding) for f_name in sorted(resolved_names)],
            n_partitions,
        )
        rdd = rdd_filenames.map(map_whole_text_file)
        rdd._name = path
        return rdd


class DummyPool(object):
    def __init__(self):
        pass

    def map(self, f, input_list):
        return (f(x) for x in input_list)

    def close(self):
        pass

    def shutdown(self):
        pass


# pickle-able helpers

def map_whole_text_file(f_name__encoding):
    f_name, encoding = f_name__encoding
    return (
        f_name,
        TextFile(f_name).load(encoding=encoding).read(),
    )


class FixedLengthChunker(object):
    def __init__(self, recordLength):
        self.record_length = recordLength

    def __call__(self, data):
        for i in range(0, len(data), self.record_length):
            yield data[i: i + self.record_length]


class VariableLengthChunker(object):
    def __init__(self, recordLength):
        self.length_fmt = recordLength
        self.prefix_length = struct.calcsize(recordLength)

    def __call__(self, data):
        while data:
            prefix, data = data[:self.prefix_length], data[self.prefix_length:]
            length = struct.unpack(self.length_fmt, prefix)[0]
            package, data = data[:length], data[length:]
            yield package



from __future__ import print_function

import logging
import unittest

import pysparkling


class Context(unittest.TestCase):
    def test_broadcast(self):
        b = pysparkling.Context().broadcast([1, 2, 3])
        self.assertEqual(b.value[0], 1)

    def test_lock1(self):
        """Should not be able to create a new RDD inside a map operation."""
        sc = pysparkling.Context()
        self.assertRaises(
            pysparkling.exceptions.ContextIsLockedException,
            lambda: (sc
                     .parallelize(range(5))
                     .map(lambda _: sc.parallelize([1]))
                     .collect())
        )

    def test_lock2(self):
        """Should not be able to create RDDs containing RDDs."""
        sc = pysparkling.Context()

        def parallelize_in_parallelize():
            o = sc.parallelize(sc.parallelize(range(x)) for x in range(5))
            print(o.map(lambda x: x.collect()).collect())

        self.assertRaises(
            pysparkling.exceptions.ContextIsLockedException,
            parallelize_in_parallelize
        )

    def test_parallelize_single_element(self):
        my_rdd = pysparkling.Context().parallelize([7], 100)
        self.assertEqual(my_rdd.collect(), [7])

    def test_parallelize_matched_elements(self):
        my_rdd = pysparkling.Context().parallelize([1, 2, 3, 4, 5], 5)
        self.assertEqual(my_rdd.collect(), [1, 2, 3, 4, 5])

    def test_parallelize_empty_partitions_at_end(self):
        my_rdd = pysparkling.Context().parallelize(range(3529), 500)
        print(my_rdd.getNumPartitions())
        my_rdd.foreachPartition(lambda p: print(sum(1 for _ in p)))
        self.assertEqual(my_rdd.getNumPartitions(), 500)
        self.assertEqual(my_rdd.count(), 3529)

    def test_retry(self):

        class EverySecondCallFails(object):
            def __init__(self):
                self.attempt = 0

            def __call__(self, value):
                self.attempt += 1
                if self.attempt % 2 == 1:
                    raise Exception
                return value

        data = list(range(6))
        rdd = pysparkling.Context().parallelize(data, 3)
        result = rdd.mapPartitions(EverySecondCallFails()).collect()
        self.assertEqual(result, data)

    def test_union(self):
        sc = pysparkling.Context()
        rdd1 = sc.parallelize(['Hello'])
        rdd2 = sc.parallelize(['World'])
        union = sc.union([rdd1, rdd2]).collect()
        print(union)
        self.assertEqual(union, ['Hello', 'World'])

    def test_version(self):
        self.assertTrue(isinstance(pysparkling.Context().version, str))


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    Context().test_retry()



import logging

log = logging.getLogger(__name__)


class FileSystem(object):
    """Interface class for the file system.

    :param str file_name: File name.
    """
    def __init__(self, file_name):
        self.file_name = file_name

    @staticmethod
    def resolve_filenames(expr):
        """Resolve the given glob-like expression to filenames.

        :rtype: list
        """
        log.error('Cannot resolve: {0}'.format(expr))

    def exists(self):
        """Check whether the given file_name exists.

        :rtype: bool
        """
        log.warning('Could not determine whether {0} exists due to '
                    'unhandled scheme.'.format(self.file_name))

    def load(self):
        """Load a file to a stream.

        :rtype: io.BytesIO
        """
        log.error('Cannot load: {0}'.format(self.file_name))

    def load_text(self, encoding='utf8', encoding_errors='ignore'):
        """Load a file to a stream.

        :param str encoding: Text encoding.
        :param str encoding_errors: How to handle encoding errors.

        :rtype: io.StringIO
        """
        log.error('Cannot load: {0}'.format(self.file_name))

    def dump(self, stream):
        """Dump a stream to a file.

        :param io.BytesIO stream: Input tream.
        """
        log.error('Cannot dump: {0}'.format(self.file_name))

    def make_public(self, recursive=False):
        """Make the file public (only on some file systems).

        :param bool recursive: Recurse.
        :rtype: FileSystem
        """
        log.warning('Cannot make {0} public.'.format(self.file_name))



from __future__ import print_function

import logging
import pysparkling


def test_trivial_sample():
    rdd = pysparkling.Context().parallelize(range(1000), 1000)
    sampled = rdd.sample(False, 0.01, 42).collect()
    print(sampled)
    assert sampled == [97, 164, 294, 695, 807, 864, 911]


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    test_trivial_sample()



from __future__ import absolute_import, unicode_literals

from fnmatch import fnmatch
import io
import logging
import os

from ...utils import Tokenizer
from .file_system import FileSystem

log = logging.getLogger(__name__)


class Local(FileSystem):
    """:class:`.FileSystem` implementation for the local file system."""

    @staticmethod
    def resolve_filenames(expr):
        if expr.startswith('file://'):
            expr = expr[7:]
        if os.path.isfile(expr):
            return [expr]
        if '/' not in expr:
            expr = '.' + os.path.sep + expr

        t = Tokenizer(expr)
        prefix = t.next(['*', '?'])
        if not prefix.endswith('/') and '/' in prefix:
            prefix = os.path.dirname(prefix)
        files = []
        for root, _, filenames in os.walk(prefix):
            for filename in filenames:
                path = os.path.join(root, filename)
                if fnmatch(path, expr) or fnmatch(path, expr + '/part*'):
                    files.append(path)
        return files

    @property
    def file_path(self):
        if self.file_name.startswith('file://'):
            return self.file_name[7:]
        return self.file_name

    def exists(self):
        return os.path.exists(self.file_path)

    def load(self):
        with io.open(self.file_path, 'rb') as f:
            return io.BytesIO(f.read())

    def load_text(self, encoding='utf8', encoding_errors='ignore'):
        with io.open(self.file_path, 'r',
                     encoding=encoding, errors=encoding_errors) as f:
            return io.StringIO(f.read())

    def dump(self, stream):
        file_path = self.file_path  # caching

        # making sure directory exists
        dirname = os.path.dirname(file_path)
        if dirname and not os.path.exists(dirname):
            log.debug('creating local directory {0}'.format(dirname))
            os.makedirs(dirname)

        log.debug('writing file {0}'.format(file_path))
        with io.open(file_path, 'wb') as f:
            for c in stream:
                f.write(c)
        return self



from __future__ import print_function

import logging
import os
import pickle
import random
import sys
import tempfile
import unittest

import pytest

from pysparkling import Context
from pysparkling.fileio import File

try:
    import py7zlib
except ImportError:
    py7zlib = None

AWS_ACCESS_KEY_ID = os.getenv('AWS_ACCESS_KEY_ID')
S3_TEST_PATH = os.getenv('S3_TEST_PATH')
OAUTH2_CLIENT_ID = os.getenv('OAUTH2_CLIENT_ID')
GS_TEST_PATH = os.getenv('GS_TEST_PATH')
HDFS_TEST_PATH = os.getenv('HDFS_TEST_PATH')
LOCAL_TEST_PATH = os.path.dirname(__file__)


def test_cache():
    # this crashes in version 0.2.28
    lines = Context().textFile('{}/*textFil*.py'.format(LOCAL_TEST_PATH))
    lines = lines.map(lambda l: '-' + l).cache()
    print(len(lines.collect()))
    lines = lines.map(lambda l: '+' + l)
    lines = lines.map(lambda l: '-' + l).cache()
    lines = lines.collect()
    print(lines)
    assert '-+-from pysparkling import Context' in lines


def test_local_textFile_1():
    lines = Context().textFile('{}/*textFil*.py'.format(LOCAL_TEST_PATH))
    lines = lines.collect()
    print(lines)
    assert 'from pysparkling import Context' in lines


def test_local_textFile_2():
    line_count = Context().textFile('{}/*.py'.format(LOCAL_TEST_PATH)).count()
    print(line_count)
    assert line_count > 90


def test_local_textFile_name():
    name = Context().textFile('{}/*.py'.format(LOCAL_TEST_PATH)).name()
    print(name)
    assert name.startswith('{}/*.py'.format(LOCAL_TEST_PATH))


def test_wholeTextFiles():
    all_files = Context().wholeTextFiles('{}/*.py'.format(LOCAL_TEST_PATH))
    this_file = all_files.lookup(__file__)
    print(this_file)
    assert 'test_wholeTextFiles' in this_file[0]


@pytest.mark.skipif(not AWS_ACCESS_KEY_ID, reason='no AWS env')
def test_s3_textFile():
    myrdd = Context().textFile(
        's3n://aws-publicdatasets/common-crawl/crawl-data/'
        'CC-MAIN-2015-11/warc.paths.*'
    )
    assert (
        'common-crawl/crawl-data/CC-MAIN-2015-11/segments/1424937481488.49/'
        'warc/CC-MAIN-20150226075801-00329-ip-10-28-5-156.ec2.'
        'internal.warc.gz' in myrdd.collect()
    )


@pytest.mark.skipif(not AWS_ACCESS_KEY_ID, reason='no AWS env')
def test_s3_textFile_loop():
    random.seed()

    fn = '{}/pysparkling_test_{:d}.txt'.format(
        S3_TEST_PATH, random.random() * 999999.0
    )

    rdd = Context().parallelize('Line {0}'.format(n) for n in range(200))
    rdd.saveAsTextFile(fn)
    rdd_check = Context().textFile(fn)

    assert (
        rdd.count() == rdd_check.count() and
        all(e1 == e2 for e1, e2 in zip(rdd.collect(), rdd_check.collect()))
    )


@pytest.mark.skipif(not HDFS_TEST_PATH, reason='no HDFS env')
def test_hdfs_textFile_loop():
    random.seed()

    fn = '{}/pysparkling_test_{:d}.txt'.format(
        HDFS_TEST_PATH, random.random() * 999999.0)
    print('HDFS test file: {0}'.format(fn))

    rdd = Context().parallelize('Hello World {0}'.format(x) for x in range(10))
    rdd.saveAsTextFile(fn)
    read_rdd = Context().textFile(fn)
    print(rdd.collect())
    print(read_rdd.collect())
    assert (
        rdd.count() == read_rdd.count() and
        all(r1 == r2 for r1, r2 in zip(rdd.collect(), read_rdd.collect()))
    )


@pytest.mark.skipif(not HDFS_TEST_PATH, reason='no HDFS env')
def test_hdfs_file_exists():
    random.seed()

    fn1 = '{}/pysparkling_test_{:d}.txt'.format(
        HDFS_TEST_PATH, random.random() * 999999.0)
    fn2 = '{}/pysparkling_test_{:d}.txt'.format(
        HDFS_TEST_PATH, random.random() * 999999.0)

    rdd = Context().parallelize('Hello World {0}'.format(x) for x in range(10))
    rdd.saveAsTextFile(fn1)

    assert File(fn1).exists() and not File(fn2).exists()


@pytest.mark.skipif(not GS_TEST_PATH, reason='no GS env')
@pytest.mark.skipif(not OAUTH2_CLIENT_ID, reason='no OAUTH env')
def test_gs_textFile_loop():
    random.seed()

    fn = '{}/pysparkling_test_{:d}.txt'.format(
        GS_TEST_PATH, random.random() * 999999.0)

    rdd = Context().parallelize('Line {0}'.format(n) for n in range(200))
    rdd.saveAsTextFile(fn)
    rdd_check = Context().textFile(fn)

    assert (
        rdd.count() == rdd_check.count() and
        all(e1 == e2 for e1, e2 in zip(rdd.collect(), rdd_check.collect()))
    )


@pytest.mark.skipif(not AWS_ACCESS_KEY_ID, reason='no AWS env')
@pytest.mark.skipif(not S3_TEST_PATH, reason='no S3 env')
def test_dumpToFile():
    random.seed()

    fn = '{}/pysparkling_test_{:d}.pickle'.format(
        S3_TEST_PATH, random.random() * 999999.0)
    File(fn).dump(pickle.dumps({'hello': 'world'}))


def test_http_textFile():
    myrdd = Context().textFile(
        'https://s3-us-west-2.amazonaws.com/human-microbiome-project/DEMO/'
        'HM16STR/46333/by_subject/1139.fsa'
    )
    assert u'TGCTGCGGTGAATGCGTTCCCGGGTCT' in myrdd.collect()


def test_saveAsTextFile():
    tempFile = tempfile.NamedTemporaryFile(delete=True)
    tempFile.close()
    Context().parallelize(range(10)).saveAsTextFile(tempFile.name)
    with open(tempFile.name, 'r') as f:
        r = f.readlines()
        print(r)
        assert '5\n' in r


def test_saveAsTextFile_tar():
    tempFile = tempfile.NamedTemporaryFile(delete=True)
    tempFile.close()
    Context().parallelize(range(10)).saveAsTextFile(tempFile.name + '.tar')
    read_rdd = Context().textFile(tempFile.name + '.tar')
    print(read_rdd.collect())
    assert '5' in read_rdd.collect()


@unittest.skipIf(hasattr(sys, 'pypy_version_info'), 'skip on pypy')
def test_saveAsTextFile_targz():
    tempFile = tempfile.NamedTemporaryFile(delete=True)
    tempFile.close()
    Context().parallelize(range(10)).saveAsTextFile(tempFile.name + '.tar.gz')
    read_rdd = Context().textFile(tempFile.name + '.tar.gz')
    print(read_rdd.collect())
    assert '5' in read_rdd.collect()


def test_saveAsTextFile_tarbz2():
    tempFile = tempfile.NamedTemporaryFile(delete=True)
    tempFile.close()
    Context().parallelize(range(10)).saveAsTextFile(tempFile.name + '.tar.bz2')
    read_rdd = Context().textFile(tempFile.name + '.tar.bz2')
    print(read_rdd.collect())
    assert '5' in read_rdd.collect()


def test_saveAsTextFile_gz():
    tempFile = tempfile.NamedTemporaryFile(delete=True)
    tempFile.close()
    Context().parallelize(range(10)).saveAsTextFile(tempFile.name + '.gz')
    read_rdd = Context().textFile(tempFile.name + '.gz')
    assert '5' in read_rdd.collect()


def test_saveAsTextFile_zip():
    tempFile = tempfile.NamedTemporaryFile(delete=True)
    tempFile.close()
    Context().parallelize(range(10)).saveAsTextFile(tempFile.name + '.zip')
    read_rdd = Context().textFile(tempFile.name + '.zip')
    print(read_rdd.collect())
    assert '5' in read_rdd.collect()


def test_saveAsTextFile_bz2():
    tempFile = tempfile.NamedTemporaryFile(delete=True)
    tempFile.close()
    Context().parallelize(range(10)).saveAsTextFile(tempFile.name + '.bz2')
    read_rdd = Context().textFile(tempFile.name + '.bz2')
    assert '5' in read_rdd.collect()


def test_saveAsTextFile_lzma():
    tempFile = tempfile.NamedTemporaryFile(delete=True)
    tempFile.close()
    Context().parallelize(range(10)).saveAsTextFile(tempFile.name + '.lzma')
    read_rdd = Context().textFile(tempFile.name + '.lzma')
    assert '5' in read_rdd.collect()


@unittest.skipIf(py7zlib is None,
                 'py7zlib import failed, is pylzma installed?')
def test_read_7z():
    # file was created with:
    # 7z a tests/data.7z tests/readme_example.py
    # (brew install p7zip)
    rdd = Context().textFile('{}/data.7z'.format(LOCAL_TEST_PATH))
    print(rdd.collect())
    assert 'from pysparkling import Context' in rdd.collect()


def test_read_tar_gz():
    # file was created with:
    # tar -cvzf data.tar.gz hello.txt
    rdd = Context().textFile('{}/data.tar.gz'.format(LOCAL_TEST_PATH))
    print(rdd.collect())
    assert 'Hello pysparkling!' in rdd.collect()


@unittest.skipIf(os.getenv('TRAVIS') is not None,
                 'skip 20news test on Travis')
def test_read_tar_gz_20news():
    # 20 news dataset has some '0xff' characters that lead to encoding
    # errors before. Adding this as a test case.
    src = 'http://qwone.com/~jason/20Newsgroups/20news-19997.tar.gz'
    rdd = Context().textFile(src, use_unicode=False)
    assert '}|> 1. Mechanical driven odometer:' in rdd.top(500)


def test_pyspark_compatibility_txt():
    kv = Context().textFile(
        '{}/pyspark/key_value.txt'.format(LOCAL_TEST_PATH)).collect()
    print(kv)
    assert u"('a', 1)" in kv and u"('b', 2)" in kv and len(kv) == 2


def test_pyspark_compatibility_bz2():
    kv = Context().textFile(
        '{}/pyspark/key_value.txt.bz2'.format(LOCAL_TEST_PATH)).collect()
    print(kv)
    assert u"a\t1" in kv and u"b\t2" in kv and len(kv) == 2


def test_pyspark_compatibility_gz():
    kv = Context().textFile(
        '{}/pyspark/key_value.txt.gz'.format(LOCAL_TEST_PATH)).collect()
    print(kv)
    assert u"a\t1" in kv and u"b\t2" in kv and len(kv) == 2


def test_local_regex_read():
    # was not working before 0.3.19
    tempFile = tempfile.NamedTemporaryFile(delete=True)
    tempFile.close()
    Context().parallelize(range(30), 30).saveAsTextFile(tempFile.name)
    d = Context().textFile(tempFile.name + '/part-0000*').collect()
    print(d)
    assert len(d) == 10


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    test_local_regex_read()



from __future__ import absolute_import

import logging
import struct
from tornado.gen import coroutine, moment
from tornado.iostream import StreamClosedError
from tornado.tcpserver import TCPServer

from ..rdd import EmptyRDD

log = logging.getLogger(__name__)


class TCPDeserializer(object):
    def __init__(self, context):
        self.context = context

    def __call__(self, data):
        if data is None:
            return EmptyRDD(self.context)

        return self.context.parallelize(data)


class TCPTextStream(TCPServer):
    def __init__(self, delimiter=b'\n'):
        super(TCPTextStream, self).__init__()
        self.delimiter = delimiter
        self.buffer = []

    def get(self):
        if not self.buffer:
            return []

        buffer_ = self.buffer
        self.buffer = []
        return buffer_

    @coroutine
    def handle_stream(self, stream, address):
        try:
            while True:
                for _ in range(100):
                    data = yield stream.read_until(self.delimiter)
                    self.buffer.append(data[:-1].decode('utf8'))
                yield moment
        except StreamClosedError:
            pass


class TCPBinaryStream(TCPServer):
    """Consumes binary messages from a TCP socket.

    :param length: An int or string.
    """

    def __init__(self, length=None):
        super(TCPBinaryStream, self).__init__()
        self.length = length
        self.buffer = []

        self.prefix_length = None
        if not isinstance(self.length, int):
            self.prefix_length = struct.calcsize(self.length)

    def get(self):
        if not self.buffer:
            return []

        buffer_ = self.buffer
        self.buffer = []
        return buffer_

    @coroutine
    def handle_stream(self, stream, address):
        try:
            while True:
                for _ in range(100):
                    if self.prefix_length:
                        prefix = yield stream.read_bytes(self.prefix_length)
                        message_length = struct.unpack(self.length, prefix)[0]
                    else:
                        message_length = self.length
                    data = yield stream.read_bytes(message_length)
                    self.buffer.append(data)
                yield moment
        except StreamClosedError:
            return



from __future__ import absolute_import

from fnmatch import fnmatch
import logging
from io import BytesIO, StringIO

from ...exceptions import FileSystemNotSupported
from ...utils import Tokenizer
from .file_system import FileSystem

log = logging.getLogger(__name__)

try:
    from gcloud import storage
except ImportError:
    storage = None


class GS(FileSystem):
    """:class:`.FileSystem` implementation for Google Storage.

    Paths are of the form `gs://bucket_name/file_path` or
    `gs://project_name:bucket_name/file_path`.
    """

    #: Set a default project name.
    project_name = None

    #: Default mime type.
    mime_type = 'text/plain'

    _clients = {}

    def __init__(self, file_name):
        if storage is None:
            raise FileSystemNotSupported(
                'Google Storage is not supported. Install "gcloud".'
            )

        super(GS, self).__init__(file_name)

        # obtain key
        t = Tokenizer(self.file_name)
        t.next('://')  # skip scheme
        bucket_name = t.next('/')
        if ':' in bucket_name:
            project_name, _, bucket_name = bucket_name.partition(':')
        else:
            project_name = GS.project_name
        blob_name = t.next()

        client = GS._get_client(project_name)
        bucket = client.get_bucket(bucket_name)
        self.blob = bucket.get_blob(blob_name)
        if not self.blob:
            self.blob = bucket.blob(blob_name)

    @staticmethod
    def _get_client(project_name):
        if project_name not in GS._clients:
            GS._clients[project_name] = storage.Client(project_name)
        return GS._clients[project_name]

    @staticmethod
    def resolve_filenames(expr):
        files = []

        t = Tokenizer(expr)
        scheme = t.next('://')
        bucket_name = t.next('/')
        if ':' in bucket_name:
            project_name, _, bucket_name = bucket_name.partition(':')
        else:
            project_name = GS.project_name
        prefix = t.next(['*', '?'])

        bucket = GS._get_client(project_name).get_bucket(bucket_name)
        expr_s = len(scheme) + 3 + len(project_name) + 1 + len(bucket_name) + 1
        expr = expr[expr_s:]
        for k in bucket.list_blobs(prefix=prefix):
            if fnmatch(k.name, expr) or fnmatch(k.name, expr + '/part*'):
                files.append('{0}://{1}:{2}/{3}'.format(
                    scheme, project_name, bucket_name, k.name))
        return files

    def exists(self):
        t = Tokenizer(self.file_name)
        t.next('//')  # skip scheme
        bucket_name = t.next('/')
        if ':' in bucket_name:
            project_name, _, bucket_name = bucket_name.partition(':')
        else:
            project_name = GS.project_name
        blob_name = t.next()
        bucket = GS._get_client(project_name).get_bucket(bucket_name)
        return (bucket.get_blob(blob_name) or
                list(bucket.list_blobs(prefix='{}/'.format(blob_name))))

    def load(self):
        log.debug('Loading {0} with size {1}.'
                  ''.format(self.blob.name, self.blob.size))
        return BytesIO(self.blob.download_as_string())

    def load_text(self, encoding='utf8', encoding_errors='ignore'):
        log.debug('Loading {0} with size {1}.'
                  ''.format(self.blob.name, self.blob.size))
        return StringIO(
            self.blob.download_as_string().decode(
                encoding, encoding_errors
            )
        )

    def dump(self, stream):
        log.debug('Dumping to {0}.'.format(self.blob.name))
        self.blob.upload_from_string(stream.read(),
                                     content_type=self.mime_type)
        return self

    def make_public(self, recursive=False):
        self.blob.make_public(recursive)
        return self



"""pysparkling module"""
# flake8: noqa

__version__ = '0.5.0'

from .rdd import RDD
from .context import Context
from .broadcast import Broadcast
from .stat_counter import StatCounter
from .cache_manager import CacheManager, TimedCacheManager

from . import fileio
from . import streaming
from . import exceptions

__all__ = ['RDD', 'Context', 'Broadcast', 'StatCounter', 'CacheManager',
           'TimedCacheManager',
           'exceptions', 'fileio', 'streaming']



from __future__ import division, print_function

from concurrent import futures
import logging
import math
import multiprocessing
import os
import pickle
import platform
import pprint
import random
import time
import timeit
import unittest

import cloudpickle

import pysparkling


class Processor(object):
    """This modifies lines but also keeps track whether it was executed."""
    def __init__(self):
        self.executed = False

    def indent_line(self, line):
        self.executed = True
        return '--- {}'.format(line)


class LazyTestInjection(object):
    def lazy_execution_test(self):
        r = self.sc.textFile(__file__)  # pylint: disable=no-member

        processor = Processor()

        r = r.map(processor.indent_line)
        self.assertFalse(processor.executed)  # pylint: disable=no-member
        r = r.map(processor.indent_line).cache()
        self.assertFalse(processor.executed)  # pylint: disable=no-member
        r = r.map(processor.indent_line)
        r.collect()
        self.assertTrue(processor.executed)  # pylint: disable=no-member


class Multiprocessing(unittest.TestCase):
    def setUp(self):
        pool = multiprocessing.Pool(4)
        self.sc = pysparkling.Context(pool=pool,
                                      serializer=cloudpickle.dumps,
                                      deserializer=pickle.loads)

    def test_basic(self):
        my_rdd = self.sc.parallelize([1, 3, 4])
        r = my_rdd.map(lambda x: x ** 2).collect()
        self.assertIn(16, r)

    def test_first(self):
        my_rdd = self.sc.parallelize([1, 2, 2, 4, 1, 3, 5, 9], 3)
        self.assertEqual(my_rdd.first(), 1)


def square_op(x):
    return x ** 2


class MultiprocessingWithoutCloudpickle(unittest.TestCase):
    def setUp(self):
        pool = multiprocessing.Pool(4)
        self.sc = pysparkling.Context(pool=pool)

    def test_basic(self):
        my_rdd = self.sc.parallelize([1, 3, 4])
        r = my_rdd.map(square_op).collect()
        self.assertIn(16, r)


class NotParallel(unittest.TestCase, LazyTestInjection):
    """Test cases in the spirit of the parallel test cases for reference."""

    def setUp(self):
        self.sc = pysparkling.Context()


class ThreadPool(unittest.TestCase, LazyTestInjection):
    def setUp(self):
        self.pool = futures.ThreadPoolExecutor(4)
        self.sc = pysparkling.Context(pool=self.pool)

    def tearDown(self):
        self.pool.shutdown()

    def test_basic(self):
        r = self.sc.parallelize([1, 3, 4]).map(math.sqrt).collect()
        self.assertIn(2, r)


class ProcessPool(unittest.TestCase):  # cannot work here: LazyTestInjection):
    def setUp(self):
        self.pool = futures.ProcessPoolExecutor(4)
        self.sc = pysparkling.Context(pool=self.pool,
                                      serializer=cloudpickle.dumps,
                                      deserializer=pickle.loads)

    def tearDown(self):
        self.pool.shutdown()

    def test_basic(self):
        r = self.sc.parallelize([1, 3, 4]).map(math.sqrt).collect()
        self.assertIn(2, r)

    def test_zipWithIndex(self):
        """Prevent regression in zipWithIndex().

        Test the case of parallelizing data directly form toLocalIterator()
        in the multiprocessing case.
        """
        r = (self.sc
             .parallelize([1, 3, 4, 9, 15, 25, 50, 75, 100], 3)
             .zipWithIndex()
             .collect())
        self.assertIn((4, 2), r)

    def test_cache(self):
        r = self.sc.parallelize(range(3), 3)

        def sleep05(v):
            time.sleep(0.5)
            return v

        r = r.map(sleep05).cache()
        self.assertEqual(r.collect(), [0, 1, 2])

        start = time.time()
        r.collect()
        self.assertLess(time.time() - start, 0.5)


class ProcessPoolIdlePerformance(unittest.TestCase):
    """Idle performance tests.

    The "load" on these tests are sleeps.
    """

    def runtime(self, n=10, processes=1):
        start = time.time()
        with futures.ProcessPoolExecutor(processes) as pool:
            sc = pysparkling.Context(pool=pool,
                                     serializer=cloudpickle.dumps,
                                     deserializer=pickle.loads)
            rdd = sc.parallelize(range(n), 10)
            rdd.map(lambda _: time.sleep(0.1)).collect()
        return time.time() - start

    @unittest.skipIf(platform.python_implementation() == 'PyPy',
                     'test fails in PyPy')
    def test_basic(self):
        t1 = self.runtime(processes=1)
        t10 = self.runtime(processes=10)
        self.assertLess(t10, t1 / 2.0)


# pickle-able map function
def map1(ft):
    return [random.choice(ft[1].split()) for _ in range(1000)]


def map_pi(n):
    return sum((
        1 for x in (random.random() ** 2 + random.random() ** 2
                    for _ in range(n))
        if x < 1.0
    ))


@unittest.skipIf(os.getenv('PERFORMANCE') is None,
                 'PERFORMANCE env variable not set')
def test_performance():
    # not pickle-able map function
    # def map2(ft):
    #     return [random.choice(ft[1].split()) for _ in range(1000)]

    def create_context(n_processes=0):
        if not n_processes:
            return pysparkling.Context()

        pool = futures.ProcessPoolExecutor(n_processes)
        return pysparkling.Context(pool=pool,
                                   serializer=cloudpickle.dumps,
                                   # serializer=pickle.dumps,
                                   deserializer=pickle.loads)

    def test(n_processes):
        sc = create_context(n_processes)
        timed = timeit.Timer(
            lambda: sc.parallelize(
                [1000 for _ in range(100)],
                100,
            ).map(map_pi).collect()
        ).timeit(number=10)
        return (timed, sc._stats)

    print('starting processing')
    n_cpu = multiprocessing.cpu_count()
    test_results = {}
    for n in range(int(n_cpu * 1.5 + 1)):
        test_results[n] = test(n)
        print(n, test_results[n][0])
    print('results where running on one core with full serialization is 1.0:')
    pprint.pprint({
        n: 1.0 / (v[0] / test_results[1][0]) for n, v in test_results.items()
    })
    print('time spent where:')
    pprint.pprint({
        n: {k: '{:.1%}'.format(t / v[1]['map_exec']) for k, t in v[1].items()}
        for n, v in test_results.items()
    })

    return (n_cpu, test_results)


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)
    # test_performance()
    t = ProcessPool()
    t.setUp()
    t.test_cache()
    t.tearDown()



from __future__ import print_function

import tornado.testing

import pysparkling


class TestCount(tornado.testing.AsyncTestCase):

    def test_count(self):
        sc = pysparkling.Context()
        ssc = pysparkling.streaming.StreamingContext(sc, 0.1)

        result = []
        (
            ssc.queueStream([range(20), ['a', 'b'], ['c']])
            .count()
            .foreachRDD(lambda rdd: result.append(rdd.collect()[0]))
        )

        ssc.start()
        ssc.awaitTermination(timeout=0.3)
        self.assertEqual(sum(result), 23)

    def test_groupByKey(self):
        sc = pysparkling.Context()
        ssc = pysparkling.streaming.StreamingContext(sc, 0.1)

        result = []
        (
            ssc.queueStream([[('a', 5), ('b', 8), ('a', 2)],
                             [('a', 2), ('b', 3)]])
            .groupByKey().mapPartitions(sorted).mapValues(sorted)
            .foreachRDD(lambda rdd: result.append(rdd.collect()))
        )

        ssc.start()
        ssc.awaitTermination(timeout=0.25)
        self.assertEqual(
            result, [[('a', [2, 5]), ('b', [8])], [('a', [2]), ('b', [3])]])

    def test_mapValues(self):
        sc = pysparkling.Context()
        ssc = pysparkling.streaming.StreamingContext(sc, 0.1)

        result = []
        (
            ssc.queueStream([[('a', [5, 8, 2]), ('b', [6, 3, 8])]])
            .mapValues(sorted)
            .foreachRDD(lambda rdd: result.append(rdd.collect()))
        )

        ssc.start()
        ssc.awaitTermination(timeout=0.15)
        self.assertEqual(result, [[('a', [2, 5, 8]), ('b', [3, 6, 8])]])



from __future__ import absolute_import, unicode_literals

from fnmatch import fnmatch
import logging
from io import BytesIO, StringIO

from ...exceptions import FileSystemNotSupported
from ...utils import parse_file_uri, format_file_uri
from .file_system import FileSystem

log = logging.getLogger(__name__)

try:
    import hdfs
except ImportError:
    hdfs = None


class Hdfs(FileSystem):
    """:class:`.FileSystem` implementation for HDFS."""

    _conn = {}

    def __init__(self, file_name):
        if hdfs is None:
            raise FileSystemNotSupported(
                'hdfs not supported. Install the python package "hdfs".'
            )

        super(Hdfs, self).__init__(file_name)

    @staticmethod
    def client_and_path(path):
        _, domain, folder_path, file_pattern = parse_file_uri(path)

        if ':' not in domain:
            port = 50070
        else:
            domain, port = domain.split(':')
            port = int(port)
        cache_id = domain + '__' + str(port)

        if cache_id not in Hdfs._conn:
            Hdfs._conn[cache_id] = hdfs.InsecureClient(  # pylint: disable=no-member
                'http://{0}:{1}'.format(domain, port)
            )
        return Hdfs._conn[cache_id], folder_path + file_pattern

    def exists(self):
        c, p = Hdfs.client_and_path(self.file_name)
        try:
            c.status(p)
        except hdfs.util.HdfsError:  # pylint: disable=no-member
            return False
        return True

    @staticmethod
    def resolve_filenames(expr):
        c, _ = Hdfs.client_and_path(expr)

        scheme, domain, folder_path, _ = parse_file_uri(expr)

        files = []
        for fn, file_status in c.list(folder_path, status=True):
            file_local_path = '{0}{1}'.format(folder_path, fn)
            file_path = format_file_uri(scheme, domain, file_local_path)
            part_file_expr = expr + ("" if expr.endswith("/") else "/") + 'part*'

            if fnmatch(file_path, expr):
                if file_status["type"] != "DIRECTORY":
                    files.append(file_path)
                else:
                    files += Hdfs._get_folder_part_files(
                        c,
                        scheme,
                        domain,
                        file_local_path,
                        part_file_expr
                    )
            elif fnmatch(file_path, part_file_expr):
                files.append(file_path)
        return files

    @staticmethod
    def _get_folder_part_files(c, scheme, domain, folder_local_path, expr_with_part):
        files = []
        for fn, file_status in c.list(folder_local_path, status=True):
            sub_file_path = format_file_uri(scheme, domain, folder_local_path, fn)
            if fnmatch(sub_file_path, expr_with_part) and file_status["type"] != "DIRECTORY":
                files.append(sub_file_path)
        return files

    def load(self):
        log.debug('Hdfs read for {0}.'.format(self.file_name))
        c, path = Hdfs.client_and_path(self.file_name)

        with c.read(path) as reader:
            r = BytesIO(reader.read())

        return r

    def load_text(self, encoding='utf8', encoding_errors='ignore'):
        log.debug('Hdfs text read for {0}.'.format(self.file_name))
        c, path = Hdfs.client_and_path(self.file_name)

        with c.read(path) as reader:
            r = StringIO(reader.read().decode(encoding, encoding_errors))

        return r

    def dump(self, stream):
        log.debug('Dump to {0} with hdfs write.'.format(self.file_name))
        c, path = Hdfs.client_and_path(self.file_name)
        c.write(path, stream)
        return self



from __future__ import absolute_import, unicode_literals

import logging
from io import BytesIO, StringIO, TextIOWrapper

from . import codec
from .file import File
from .fs.file_system import FileSystem

log = logging.getLogger(__name__)

# there is no basestring in Python 3, so define string_types:
try:
    string_types = basestring
except NameError:
    string_types = str


class TextFile(File):
    """Derived from :class:`File`.

    :param file_name: Any text file name.
    """

    def load(self, encoding='utf8', encoding_errors='ignore'):  # pylint: disable=arguments-differ
        """Load the data from a file.

        :param str encoding: The character encoding of the file.
        :param str encoding_errors: How to handle encoding errors.
        :rtype: io.StringIO
        """
        if isinstance(self.codec, codec.NoCodec) and \
           self.fs.load_text != FileSystem.load_text:
            stream = self.fs.load_text(encoding, encoding_errors)
        else:
            stream = self.fs.load()
            stream = self.codec.decompress(stream)
            stream = TextIOWrapper(stream, encoding, encoding_errors)
        return stream

    def dump(self, stream=None, encoding='utf8', encoding_errors='ignore'):  # pylint: disable=arguments-differ
        """Writes a stream to a file.

        :param stream:
            An ``io.StringIO`` instance. A ``basestring`` is also possible and
            get converted to ``io.StringIO``.

        :param encoding: (optional)
            The character encoding of the file.

        :rtype: TextFile
        """
        if stream is None:
            stream = StringIO()

        if isinstance(stream, string_types):
            stream = StringIO(stream)

        stream = self.codec.compress(
            BytesIO(stream.read().encode(encoding, encoding_errors))
        )
        self.fs.dump(stream)

        return self



from __future__ import print_function

import os

import pytest

from pysparkling.fileio import File

CURRENT_FILE_LOCATION = __file__


class MockedHdfsClient(object):
    def list(self, path, status):
        if path == "/user/username/":
            return [
                ("input", {"type": "DIRECTORY"}),
                ("output", {"type": "DIRECTORY"})
            ]
        if path in ('/user/username/input', '/user/username/input/'):
            return [
                ("part-00001.gz", {"type": "FILE"}),
                ("part-00002.gz", {"type": "FILE"}),
                ("_SUCCESS", {"type": "FILE"})
            ]
        raise NotImplementedError("Return value not mocked for '{0}'".format(path))


class MockedS3Bucket(object):
    def list(self, *args, **kwargs):
        return [
            MockedS3Key("user/username/input/part-00001.gz"),
            MockedS3Key("user/username/input/part-00002.gz"),
            MockedS3Key("user/username/input/_SUCCESS"),
        ]


class MockedS3Connection(object):
    def get_bucket(self, *args, **kwargs):
        return MockedS3Bucket()


class MockedS3Key(object):
    def __init__(self, name):
        self.name = name


def test_local_1():
    filenames = File.resolve_filenames(
        '{}/*'.format(os.path.dirname(CURRENT_FILE_LOCATION))
    )
    assert CURRENT_FILE_LOCATION in filenames


def test_local_2():
    filenames = File.resolve_filenames(CURRENT_FILE_LOCATION)
    assert filenames == [CURRENT_FILE_LOCATION]


@pytest.mark.skipif(not os.getenv('AWS_ACCESS_KEY_ID'), reason='no AWS env')
def test_s3_1():
    filenames = File.resolve_filenames(
        's3n://aws-publicdatasets/common-crawl/'
        'crawl-data/CC-MAIN-2015-11/warc.paths.*'
    )
    print(filenames)
    assert ('s3n://aws-publicdatasets/common-crawl/'
            'crawl-data/CC-MAIN-2015-11/warc.paths.gz' in filenames)


def test_hdfs_resolve_filenames_with_wildcard():
    from pysparkling.fileio.fs import Hdfs
    Hdfs.client_and_path = staticmethod(lambda *args, **kwargs: (MockedHdfsClient(), "unused_path"))

    filenames = Hdfs.resolve_filenames("hdfs://hdfs-cluster.com/user/username/input/part-*.gz")
    print(filenames)
    assert filenames == [
        'hdfs://hdfs-cluster.com/user/username/input/part-00001.gz',
        'hdfs://hdfs-cluster.com/user/username/input/part-00002.gz'
    ]


def test_hdfs_resolve_filenames_with_folder_path():
    from pysparkling.fileio.fs import Hdfs
    Hdfs.client_and_path = staticmethod(lambda *args, **kwargs: (MockedHdfsClient(), "unused_path"))

    filenames = Hdfs.resolve_filenames("hdfs://hdfs-cluster.com/user/username/input")
    print(filenames)
    assert filenames == [
        'hdfs://hdfs-cluster.com/user/username/input/part-00001.gz',
        'hdfs://hdfs-cluster.com/user/username/input/part-00002.gz'
    ]


def test_hdfs_resolve_filenames_with_folder_path_and_trailing_slash():
    from pysparkling.fileio.fs import Hdfs
    Hdfs.client_and_path = staticmethod(lambda *args, **kwargs: (MockedHdfsClient(), "unused_path"))

    filenames = Hdfs.resolve_filenames("hdfs://hdfs-cluster.com/user/username/input/")
    print(filenames)
    assert filenames == [
        'hdfs://hdfs-cluster.com/user/username/input/part-00001.gz',
        'hdfs://hdfs-cluster.com/user/username/input/part-00002.gz'
    ]


def test_hdfs_resolve_filenames_with_file_path():
    from pysparkling.fileio.fs import Hdfs
    Hdfs.client_and_path = staticmethod(lambda *args, **kwargs: (MockedHdfsClient(), "unused_path"))

    filenames = Hdfs.resolve_filenames("hdfs://hdfs-cluster.com/user/username/input/part-00001.gz")
    print(filenames)
    assert filenames == [
        'hdfs://hdfs-cluster.com/user/username/input/part-00001.gz'
    ]


def test_s3_resolve_filenames():
    from pysparkling.fileio.fs import S3
    S3._get_conn = classmethod(lambda *args, **kwargs: MockedS3Connection())

    filenames = S3.resolve_filenames("s3://bucket-name/user/username/input/part-*.gz")
    print(filenames)
    assert filenames == [
        's3://bucket-name/user/username/input/part-00001.gz',
        's3://bucket-name/user/username/input/part-00002.gz'
    ]


if __name__ == '__main__':
    test_local_1()
    test_local_2()
    test_s3_1()
    test_hdfs_resolve_filenames_with_folder_path()
    test_hdfs_resolve_filenames_with_folder_path_and_trailing_slash()
    test_hdfs_resolve_filenames_with_file_path()
    test_hdfs_resolve_filenames_with_wildcard()
    test_s3_resolve_filenames()



from __future__ import absolute_import

from ..rdd import RDD, EmptyRDD


class QueueStreamDeserializer(object):
    def __init__(self, context):
        self.context = context

    def ensure_rdd(self, data):
        if data is None:
            return EmptyRDD(self.context)
        if isinstance(data, RDD):
            return data
        return self.context.parallelize(data)

    def __call__(self, data):
        return self.ensure_rdd(data)


class QueueStream(object):
    def __init__(self, queue, oneAtATime=True, default=None):
        self.queue = queue
        self.oneAtATime = oneAtATime
        self.default = default

    def get(self):
        q_size = self.queue.qsize()

        if q_size == 0:
            return self.default

        if self.oneAtATime:
            return self.queue.get_nowait()

        return [e for _ in range(q_size) for e in self.queue.get_nowait()]



import tornado.testing

import pysparkling


class TextFile(tornado.testing.AsyncTestCase):

    def test_connect(self):
        sc = pysparkling.Context()
        ssc = pysparkling.streaming.StreamingContext(sc, 0.1)

        result = []
        (
            ssc.textFileStream('LICENS*', process_all=True)
            .count()
            .foreachRDD(lambda rdd: result.append(rdd.collect()[0]))
        )

        ssc.start()
        ssc.awaitTermination(timeout=0.3)
        self.assertEqual(sum(result), 22)

    def test_save(self):
        sc = pysparkling.Context()
        ssc = pysparkling.streaming.StreamingContext(sc, 0.1)

        (
            ssc.textFileStream('LICENS*')
            .count()
            .saveAsTextFiles('tests/textout/')
        )

    def test_save_gz(self):
        sc = pysparkling.Context()
        ssc = pysparkling.streaming.StreamingContext(sc, 0.1)

        (
            ssc.textFileStream('LICENS*')
            .count()
            .saveAsTextFiles('tests/textout/', suffix='.gz')
        )


class BinaryFile(tornado.testing.AsyncTestCase):

    def test_read_file(self):
        sc = pysparkling.Context()
        ssc = pysparkling.streaming.StreamingContext(sc, 0.1)

        result = []
        (
            ssc.fileBinaryStream('LICENS*', process_all=True)
            .count()
            .foreachRDD(lambda rdd: result.append(rdd.collect()[0]))
        )

        ssc.start()
        ssc.awaitTermination(timeout=0.3)
        self.assertEqual(sum(result), 1)

    def test_read_chunks(self):
        sc = pysparkling.Context()
        ssc = pysparkling.streaming.StreamingContext(sc, 0.1)

        result = []
        (
            ssc.fileBinaryStream('LICENS*', recordLength=40, process_all=True)
            .count()
            .foreachRDD(lambda rdd: result.append(rdd.collect()[0]))
        )

        ssc.start()
        ssc.awaitTermination(timeout=0.3)
        self.assertEqual(sum(result), 28)



import unittest

from operator import add
from pysparkling import Context


class RDDTest(unittest.TestCase):
    """Tests for the resilient distributed databases"""

    def setUp(self):
        self.context = Context()

    def testLeftOuterJoinSimple(self):
        """Test the basic left outer join with simple key-value pairs"""
        x = self.context.parallelize([('a', 'xa'), ('b', 'xb'), ('c', 'xc')])
        y = self.context.parallelize([('b', 'yb'), ('c', 'yc')])
        z = self.context.parallelize([('c', 'zc'), ('d', 'zd')])

        xy = sorted(x.leftOuterJoin(y).collect())
        xz = sorted(x.leftOuterJoin(z).collect())
        zx = sorted(z.leftOuterJoin(x).collect())

        self.assertEqual(xy, [('a', ('xa', None)),
                              ('b', ('xb', 'yb')),
                              ('c', ('xc', 'yc'))])

        self.assertEqual(xz, [('a', ('xa', None)),
                              ('b', ('xb', None)),
                              ('c', ('xc', 'zc'))])

        self.assertEqual(zx, [('c', ('zc', 'xc')),
                              ('d', ('zd', None))])

    def testLeftOuterJoinDuplicate(self):
        """Test the left outer join with duplicate keys"""
        x = self.context.parallelize([('a', 'xa'), ('c', 'xc1'), ('c', 'xc2')])
        y = self.context.parallelize([('b', 'yb'), ('c', 'yc')])
        z = self.context.parallelize([('c', 'zc1'), ('c', 'zc2'), ('d', 'zd')])

        xy = sorted(x.leftOuterJoin(y).collect())
        xz = sorted(x.leftOuterJoin(z).collect())

        self.assertEqual(xy, [('a', ('xa', None)),
                              ('c', ('xc1', 'yc')),
                              ('c', ('xc2', 'yc'))])

        # Two sets of duplicate keys gives cartesian product
        self.assertEqual(xz, [('a', ('xa', None)),
                              ('c', ('xc1', 'zc1')),
                              ('c', ('xc1', 'zc2')),
                              ('c', ('xc2', 'zc1')),
                              ('c', ('xc2', 'zc2'))])

    def testRightOuterJoinSimple(self):
        """Test the basic right outer join with simple key-value pairs"""
        x = self.context.parallelize([('a', 'xa'), ('b', 'xb'), ('c', 'xc')])
        y = self.context.parallelize([('b', 'yb'), ('c', 'yc')])
        z = self.context.parallelize([('c', 'zc'), ('d', 'zd')])

        xy = sorted(x.rightOuterJoin(y).collect())
        xz = sorted(x.rightOuterJoin(z).collect())
        zx = sorted(z.rightOuterJoin(x).collect())

        self.assertEqual(xy, [('b', ('xb', 'yb')),
                              ('c', ('xc', 'yc'))])

        self.assertEqual(xz, [('c', ('xc', 'zc')),
                              ('d', (None, 'zd'))])

        self.assertEqual(zx, [('a', (None, 'xa')),
                              ('b', (None, 'xb')),
                              ('c', ('zc', 'xc'))])

    def testRightOuterJoinDuplicate(self):
        """Test the right outer join with duplicate keys"""
        x = self.context.parallelize([('a', 'xa'), ('c', 'xc1'), ('c', 'xc2')])
        y = self.context.parallelize([('b', 'yb'), ('c', 'yc')])
        z = self.context.parallelize([('c', 'zc1'), ('c', 'zc2'), ('d', 'zd')])

        xy = sorted(x.rightOuterJoin(y).collect())
        xz = sorted(x.rightOuterJoin(z).collect())

        self.assertEqual(xy, [('b', (None, 'yb')),
                              ('c', ('xc1', 'yc')),
                              ('c', ('xc2', 'yc'))])

        # Two sets of duplicate keys gives cartesian product
        self.assertEqual(xz, [('c', ('xc1', 'zc1')),
                              ('c', ('xc1', 'zc2')),
                              ('c', ('xc2', 'zc1')),
                              ('c', ('xc2', 'zc2')),
                              ('d', (None, 'zd'))])

    def testFullOuterJoinSimple(self):
        """Test the basic full outer join with simple key-value pairs"""
        x = self.context.parallelize([('a', 'xa'), ('b', 'xb'), ('c', 'xc')])
        y = self.context.parallelize([('b', 'yb'), ('c', 'yc')])
        z = self.context.parallelize([('c', 'zc'), ('d', 'zd')])

        xy = sorted(x.fullOuterJoin(y).collect())
        xz = sorted(x.fullOuterJoin(z).collect())
        zx = sorted(z.fullOuterJoin(x).collect())

        self.assertEqual(xy, [('a', ('xa', None)),
                              ('b', ('xb', 'yb')),
                              ('c', ('xc', 'yc'))])

        self.assertEqual(xz, [('a', ('xa', None)),
                              ('b', ('xb', None)),
                              ('c', ('xc', 'zc')),
                              ('d', (None, 'zd'))])

        self.assertEqual(zx, [('a', (None, 'xa')),
                              ('b', (None, 'xb')),
                              ('c', ('zc', 'xc')),
                              ('d', ('zd', None))])

    def testFullOuterJoinDuplicate(self):
        """Test the full outer join with duplicate keys"""
        x = self.context.parallelize([('a', 'xa'), ('c', 'xc1'), ('c', 'xc2')])
        y = self.context.parallelize([('b', 'yb'), ('c', 'yc')])
        z = self.context.parallelize([('c', 'zc1'), ('c', 'zc2'), ('d', 'zd')])

        xy = sorted(x.fullOuterJoin(y).collect())
        xz = sorted(x.fullOuterJoin(z).collect())

        self.assertEqual(xy, [('a', ('xa', None)),
                              ('b', (None, 'yb')),
                              ('c', ('xc1', 'yc')),
                              ('c', ('xc2', 'yc'))])

        # Two sets of duplicate keys gives cartesian product
        self.assertEqual(xz, [('a', ('xa', None)),
                              ('c', ('xc1', 'zc1')),
                              ('c', ('xc1', 'zc2')),
                              ('c', ('xc2', 'zc1')),
                              ('c', ('xc2', 'zc2')),
                              ('d', (None, 'zd'))])

    def test_cartesian(self):
        x = self.context.parallelize(range(0, 2))
        y = self.context.parallelize(range(3, 6))
        c = x.cartesian(y)
        result = sorted(c.collect())
        expected = sorted([(0, 3), (0, 4), (0, 5), (1, 3), (1, 4), (1, 5)])
        self.assertListEqual(result, expected)

    def test_sample(self):
        rdd = self.context.parallelize(range(100), 4)
        self.assertTrue(6 <= rdd.sample(False, 0.1, 81).count() <= 14)

    def test_sampleByKey(self):
        fractions = {"a": 0.2, "b": 0.1}
        range_rdd = self.context.parallelize(range(0, 1000))
        rdd = self.context.parallelize(fractions.keys()).cartesian(range_rdd)
        sample = dict(
            rdd.sampleByKey(False, fractions, 2).groupByKey().collect()
        )
        self.assertTrue(100 < len(sample["a"]) < 300 and
                        50 < len(sample["b"]) < 150)
        self.assertTrue(max(sample["a"]) <= 999 and min(sample["a"]) >= 0)
        self.assertTrue(max(sample["b"]) <= 999 and min(sample["b"]) >= 0)

    def test_groupByKey(self):
        # This will fail if the values of the RDD need to be compared
        class IncomparableValue(object):
            def __init__(self, value):
                self.value = value

            def __eq__(self, other):
                return self.value == other.value

            def __lt__(self, other):
                raise NotImplementedError("This object cannot be compared")

        keys = (0, 1, 2, 0, 1, 2)
        r = [IncomparableValue(i) for i in range(len(keys))]

        k_rdd = self.context.parallelize(zip(keys, r))
        actual_group = k_rdd.groupByKey().collect()

        expected_group = ((0, r[0::3]),
                          (1, r[1::3]),
                          (2, r[2::3]))

        grouped_dict = {k: v for k, v in actual_group}

        for k, v in expected_group:
            self.assertIn(k, grouped_dict)

            for vv in v:
                self.assertIn(vv, grouped_dict[k])

    def test_reduceByKey(self):
        # This will fail if the values of the RDD need to be compared
        class IncomparableValueAddable(object):
            def __init__(self, value):
                self.value = value

            def __eq__(self, other):
                return self.value == other.value

            def __add__(self, other):
                return self.__class__(self.value + other.value)

            def __lt__(self, other):
                raise NotImplementedError("This object cannot be compared")

        keys = (0, 1, 2, 0, 1, 2)
        r = [IncomparableValueAddable(i) for i in range(len(keys))]

        k_rdd = self.context.parallelize(zip(keys, r))
        actual_group = k_rdd.reduceByKey(add).collect()

        expected_group = ((0, IncomparableValueAddable(3)),
                          (1, IncomparableValueAddable(5)),
                          (2, IncomparableValueAddable(7)))

        grouped_dict = {k: v for k, v in actual_group}

        # Keep this order-agnostic
        for k, v in expected_group:
            self.assertEqual(grouped_dict[k], v)

    def test_reduceByKey_with_numPartition(self):
        # This will fail if the values of the RDD need to be compared
        class IncomparableValueAddable(object):
            def __init__(self, value):
                self.value = value

            def __eq__(self, other):
                return self.value == other.value

            def __add__(self, other):
                return self.__class__(self.value + other.value)

            def __lt__(self, other):
                raise NotImplementedError("This object cannot be compared")

        keys = (0, 1, 2, 0, 1, 2)
        r = [IncomparableValueAddable(i) for i in range(len(keys))]

        k_rdd = self.context.parallelize(zip(keys, r))
        actual_group = k_rdd.reduceByKey(add, numPartitions=20).collect()

        expected_group = ((0, IncomparableValueAddable(3)),
                          (1, IncomparableValueAddable(5)),
                          (2, IncomparableValueAddable(7)))

        grouped_dict = {k: v for k, v in actual_group}

        # Keep this order-agnostic
        for k, v in expected_group:
            self.assertEqual(grouped_dict[k], v)


if __name__ == "__main__":
    unittest.main()



from __future__ import absolute_import


# lzma only available in Python >= 3.3
try:
    import lzma
except ImportError:
    lzma = None

from io import BytesIO
import logging

from .codec import Codec

log = logging.getLogger(__name__)


class Lzma(Codec):
    """Implementation of :class:`.Codec` for lzma compression.

    Needs Python >= 3.3.
    """

    def __init__(self):
        if lzma is None:
            log.warning('LZMA codec not supported. It is only supported '
                        'in Python>=3.3. Not compressing streams.')
        super(Lzma, self).__init__()

    def compress(self, stream):
        if lzma is None:
            return Codec.compress(self, stream)

        return BytesIO(lzma.compress(stream.read()))

    def decompress(self, stream):
        if lzma is None:
            return Codec.decompress(self, stream)

        return BytesIO(lzma.decompress(stream.read()))
